import base64

import aiobotocore
from vmshepherd.presets import AbstractConfigurationDriver


class AwsPresetDriver(AbstractConfigurationDriver):

    def __init__(self, config, runtime, defaults):
        super().__init__(runtime, defaults)
        self._presets = {}

    async def _get_preset_spec(self, preset_name: str):
        return self._specs[preset_name]

    async def _list(self):
        await self._reload()
        return self._specs.keys()

    async def _reload(self):

        session = aiobotocore.get_session()
        async with session.create_client('autoscaling') as asg, session.create_client('ec2') as ec2:
            res = await asg.describe_auto_scaling_groups()

            _tmp_specs = {}
            for preset in res['AutoScalingGroups']:
                config = {}
                config['name'] = preset_name = preset['AutoScalingGroupName']
                config['count'] = preset['DesiredCapacity']
                config['network'] = {}
                config['network']['availability_zone'] = ','.join(preset['AvailabilityZones'])

                # preset is managed by ASG
                config['unmanaged'] = True

                if 'LaunchConfigurationName' in preset:
                    res = await asg.describe_launch_configurations(
                        LaunchConfigurationNames=[preset['LaunchConfigurationName']]
                    )
                    launch_config = res['LaunchConfigurations'][0]
                    config['flavor'] = launch_config['InstanceType']
                    config['image'] = launch_config['ImageId']
                    config['network']['security_groups'] = launch_config['SecurityGroups']
                    if 'UserData' in launch_config:
                        config['user_data'] = base64.b64decode(launch_config['UserData']).decode('utf-8')

                elif 'LaunchTemplate' in preset:
                    version = '$Latest'
                    if 'Version' in preset['LaunchTemplate']:
                        version = preset['LaunchTemplate']['Version']
                    res = await ec2.describe_launch_template_versions(
                        LaunchTemplateId=preset['LaunchTemplate']['LaunchTemplateId'],
                        Versions=[version]
                    )

                    launch_config = res['LaunchTemplateVersions'][0]['LaunchTemplateData']
                    config['flavor'] = launch_config['InstanceType']
                    config['image'] = launch_config['ImageId']
                    config['network']['security_groups'] = launch_config['SecurityGroupIds']
                    if 'UserData' in launch_config:
                        config['userdata'] = base64.b64decode(launch_config['UserData']).decode('utf-8')

                _tmp_specs[preset_name] = config
        self._specs = _tmp_specs

    def reconfigure(self, config, defaults):
        super().reconfigure(config, defaults)



import datetime
from unittest.mock import MagicMock, Mock, patch

from aiounittest import AsyncTestCase
from aiounittest.mock import AsyncMockIterator
from asynctest import CoroutineMock

from vmshepherd_aws_drivers import AwsIaaSDriver
from vmshepherd.iaas.vm import Vm

ec2_created_at_mock = datetime.datetime(1970, 1, 1, 0, 0, tzinfo=datetime.timezone.utc)

ec2_mock = {
    'InstanceId': 'someinstanceid',
    'PrivateIpAddress': '10.188.25.12',
    'LaunchTime': ec2_created_at_mock,
    'InstanceType': 't5-large',
    'ImageId': 'someimageid',
    'State': {'Name': 'running'},
    'Tags': ['tag1', 'tag2']
}

mapping_ec2_mock = Vm(
    'someinstanceid', 'someinstanceid', ['10.188.25.12'], ec2_created_at_mock, 'running', ['tag1', 'tag2'],
    ['tag1', 'tag2'], 't5-large', 'someimageid'
)


class MockSession(MagicMock):

    async def __aenter__(self, a, b):
        return self

    def __aexit__(self):
        return self


class MockAWSServices(MagicMock):

    async def __aenter__(self):
        return self

    async def __aexit__(self, a, b, c):
        return self

    def get_paginator(self, method):
        return self

    def paginate(self, PaginationConfig={}, Filters={}):
        return AsyncMockIterator([{
            'Reservations': [{'Instances': [ec2_mock]}]
        }])


class TestAwsIaaSDriver(AsyncTestCase):

    def setUp(self):
        self.aiobotocore_mock = Mock()
        self.session_mock = MockSession()
        self.aws_services_mock = MockAWSServices()
        self.session_mock.create_client.return_value = self.aws_services_mock
        self.aiobotocore_mock.get_session = MagicMock(return_value=self.session_mock)
        patch('vmshepherd_aws_drivers.aws_iaas_driver.aiobotocore', self.aiobotocore_mock).start()
        self.aws_driver = AwsIaaSDriver({})

    def test_map_vm_structure(self):
        structure = self.aws_driver._map_vm_structure(ec2_mock)
        self.assertEqual(structure, mapping_ec2_mock)

    async def test_get_vm(self):
        ec2_result = {
                'Reservations': [{
                    'Instances': [ec2_mock]
                        }]}
        self.aws_services_mock.describe_instances = CoroutineMock(return_value=ec2_result)
        vm = await self.aws_driver.get_vm('someinstanceid')
        self.assertEqual(vm, mapping_ec2_mock)

    async def test_list_vms(self):
        vms = await self.aws_driver.list_vms('somepresetname')
        self.assertEqual(vms[0], mapping_ec2_mock)



from itertools import chain
from typing import Dict, List

import aiobotocore
from botocore.exceptions import ClientError
from vmshepherd.errors import VmNotFound
from vmshepherd.iaas import AbstractIaasDriver, Vm, VmState


class AwsIaaSDriver(AbstractIaasDriver):

    _VM_STATUSES = {
         'shutting-down': VmState.TERMINATED,
         'terminated':  VmState.TERMINATED,
         'stopping':  VmState.TERMINATED,
         'stopped':  VmState.TERMINATED,
         'pending': VmState.PENDING,
         'running': VmState.RUNNING
    }

    def __init__(self, config):
        self.config = config

    async def list_vms(self, preset_name: str) -> List:
        '''
        List VMs by preset name
        :arg present_name: string
        '''
        session = aiobotocore.get_session()
        async with session.create_client('ec2') as client:
            paginator = client.get_paginator('describe_instances')
            filters = [
                {
                    'Name': 'tag:aws:autoscaling:groupName',
                    'Values': [preset_name]
                }
            ]
            instances = []
            async for result in paginator.paginate(
                PaginationConfig={'PageSize': self.config.get('ec2_page_size', 5)},
                Filters=filters
            ):
                instances = chain(instances, *[vms['Instances'] for vms in result['Reservations']])

        return [self._map_vm_structure(instance) for instance in instances]

    async def create_vm(self):
        '''
        NotImplemented - preset is managed by ASG
        '''
        raise Exception('Preset is managed by ASG')

    async def get_vm(self, vm_id: str) -> Vm:
        session = aiobotocore.get_session()
        async with session.create_client('ec2') as client:
            try:
                res = await client.describe_instances(
                    InstanceIds=[vm_id]
                )
            except ClientError as ex:
                if ex.response['Error']['Code'] == 'InvalidInstanceID.NotFound':
                    raise VmNotFound(vm_id)
                else:
                    raise
        return self._map_vm_structure(res['Reservations'][0]['Instances'][0])

    async def terminate_vm(self, vm_id: str):
        session = aiobotocore.get_session()
        async with session.create_client('ec2') as client:
            await client.terminate_instances(
                InstanceIds=[vm_id]
            )

    def _map_vm_structure(self, instance: Dict) -> Vm:
        '''
        Vm unification
        :arg instance: object
        :returns object
        '''
        ip = [instance.get('PrivateIpAddress', '')]
        created = instance['LaunchTime']
        flavor = instance['InstanceType']
        image = instance['ImageId']
        state = self._map_vm_status(instance['State']['Name'])
        iaasvm = Vm(instance['InstanceId'], instance['InstanceId'], ip, created, state=state,
                    metadata=instance.get('Tags', []), tags=instance.get('Tags', []), flavor=flavor,
                    image=image, timed_shutdown_at=None)
        return iaasvm

    def _map_vm_status(self, vm_status: str) -> str:
        '''
         Map AWS vm statuses to vmshepherd vm statuses
         AWS vm statuses: pending | running | shutting-down | terminated | stopping | stopped

         :arg string vm_status
         :returns string
        '''

        return self._VM_STATUSES.get(vm_status, VmState.UNKNOWN)



import io
import re
from os.path import abspath, dirname, join

from setuptools import find_packages, setup


def read(name):
    here = abspath(dirname(__file__))
    return io.open(
        join(here, name), encoding='utf8'
    ).read()


setup(
    name="vmshepherd-aws-drivers",
    version="1.1.0",
    author='Dreamlab - PaaS KRK',
    author_email='paas-support@dreamlab.pl',
    url='https://github.com/Dreamlab/vmshepherd-aws-drivers',
    description='AWS drivers for VmShepherd',
    long_description='%s\n%s' % (
        read('README.rst'),
        re.sub(':[a-z]+:`~?(.*?)`', r'``\1``', read('CHANGELOG.rst'))
    ),
    packages=find_packages('src'),
    package_dir={'': 'src'},
    include_package_data=True,
    install_requires=read('requirements.txt').split('\n'),
    zip_safe=False,
    entry_points={
        'vmshepherd.driver.iaas': [
            'AwsIaaSDriver = vmshepherd_aws_drivers:AwsIaaSDriver'
        ],
        'vmshepherd.driver.presets': [
            'AwsPresetDriver = vmshepherd_aws_drivers:AwsPresetDriver'
        ]
    },
    keywords=['vmshepherd', 'AWS', 'ASG'],
    classifiers=[
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Operating System :: POSIX'
    ]
)
